import json

try:
    import redis
except ImportError as e:
    raise ImportError('Failed to import "redis" ({})'.format(e))

from .abstract import Abstract
from .abstract import NotSupported

import logging

def create_client(**pool_args):
    if 'port' not in pool_args:
        pool_args['port'] = 6379

    if 'host' not in pool_args:
        raise ValueError('"host" is not defined.')

    pool = redis.ConnectionPool(**pool_args)

    return redis.Redis(connection_pool=pool)

class Adapter(Abstract):
    def __init__(self, storage = None, namespace = None, delimiter = ':', auto_json_convertion = True):
        self._storage   = storage
        self._namespace = namespace or ''
        self._delimiter = delimiter

        self._auto_json_convertion = auto_json_convertion

    def get(self, key):
        actual_key = self._actual_key(key)
        value      = self._storage.get(actual_key)

        if not value:
            return value

        if self._auto_json_convertion:
            return json.loads(value.decode('utf-8'))

        return value

    def set(self, key, value):
        actual_key = self._actual_key(key)
        encoded    = value

        if self._auto_json_convertion:
            encoded = json.dumps(value)

        self._storage.set(actual_key, encoded)

    def remove(self, key):
        actual_key = self._actual_key(key)

        self._storage.delete(actual_key)

    def find(self, pattern='*', only_keys=False):
        actual_pattern = self._actual_key(pattern)
        if only_keys:
            return [
                key.decode('utf-8')
                for key in self._storage.keys(actual_pattern)
            ]

        return {
            key.decode('utf-8'): self.get(key)
            for key in self._storage.keys(actual_pattern)
        }

    def _actual_key(self, key):
        if not self._namespace:
            return key

        return '{}{}{}'.format(self._namespace, self._delimiter, key)



import re

from .inmemory  import InMemory

try:
    from .dynamodb  import DynamoDB
    from .memcached import Memcached
    from .memcached import get_default_client as create_memcached_client
    from .xredis    import Adapter            as Redis
    from .xredis    import create_client      as create_redis_client
except ImportError as e:
    if not re.search("No module named", str(e)):
        raise e



import json

try:
    from pymemcache.client.base import Client
    from pymemcache.client.hash import HashClient
except ImportError as e:
    raise ImportError('Failed to import "pymemcache" ({})'.format(e))

from .abstract import Abstract
from .abstract import NotSupported

def get_default_client(server_config = None):
    if server_config and isinstance(server_config, list):
        return HashClient(
            server_config,
            serializer   = serializer,
            deserializer = deserializer
        )

    return Client(
        server_config or ('127.0.0.1', 11211),
        serializer   = serializer,
        deserializer = deserializer
    )

def serializer(key, value):
    if type(value) == str:
        return value, 1

    return json.dumps(value), 2

def deserializer(key, value, flags):
    actual_value = value.decode('utf-8')

    if flags == 1:
        return actual_value

    if flags == 2:
        return json.loads(actual_value)

    # If all else fail, the deserializer will forcefully decode the object.
    try:
        return json.loads(actual_value)
    except Exception as e:
        return actual_value

class Memcached(Abstract):
    def __init__(self, storage = None, namespace = None, delimiter = ':'):
        self._storage   = storage
        self._namespace = namespace or ''
        self._delimiter = delimiter

    def get(self, key):
        actual_key = self._actual_key(key)

        return self._storage.get(actual_key)

    def set(self, key, value):
        actual_key = self._actual_key(key)

        self._storage.set(actual_key, value)

    def remove(self, key):
        actual_key = self._actual_key(key)

        self._storage.delete(actual_key)

    def find(self, *args, **kwargs):
        raise NotSupported()

    def _actual_key(self, key):
        if not self._namespace:
            return key

        return '{}{}{}'.format(self._namespace, self._delimiter, key)



import decimal
import json

try:
    from boto3.dynamodb.conditions import Key, Attr
except ImportError as e:
    raise ImportError('Failed to import "boto3" ({})'.format(e))

from .abstract import Abstract

class DynamoDB(Abstract):
    __primary_key     = 'identifier'; # the main identifier
    __secondary_key   = 'search_key'; # the search key (the content is the same as the PK)
    __value_attribute = 'value';

    def __init__(self, storage, namespace = 'default'):
        self._namespace = namespace
        self._storage   = storage

    def get(self, key):
        response = self._table().get_item(Key = self._get_key(key))

        item = response['Item'] if 'Item' in response else None

        return item[self.__value_attribute] if item else None

    def set(self, key, value):
        item = {
            self.__value_attribute: self._prepare_for_setter(value),
        }

        item.update(self._get_key(key))

        self._table().put_item(Item = item)

    def remove(self, key):
        self._table().delete_item(Key = self._get_key(key))

    def find(self, begins_with=None, eq=None):
        criteria = {}

        if eq:
            criteria['KeyConditionExpression'] = Key(self.__primary_key).eq(eq)
        elif begins_with:
            criteria['FilterExpression'] = Key(self.__secondary_key).begins_with(begins_with)

        if not criteria:
            raise ValueError('Need to define the criteria.')

        response = self._table().scan(**criteria)

        if 'Items' not in response:
            return {}

        return {
            item[self.__primary_key]: item[self.__value_attribute]
            for item in response['Items']
        }

    def prepare(self, io_read = 5, io_write = 4):
        throughput = {
            'ReadCapacityUnits':  io_read,
            'WriteCapacityUnits': io_write,
        }

        self._storage.create_table(
            TableName = self._table_name(),
            KeySchema = [
                {
                    'AttributeName': self.__primary_key,
                    'KeyType':       'HASH',
                },
                {
                    'AttributeName': self.__secondary_key,
                    'KeyType':       'RANGE',
                },
            ],
            AttributeDefinitions = [
                {
                    'AttributeName': self.__primary_key,
                    'AttributeType': 'S',
                },
                {
                    'AttributeName': self.__secondary_key,
                    'AttributeType': 'S',
                },
            ],
            ProvisionedThroughput = throughput
        )

    def _get_key(self, key):
        return {
            self.__primary_key:   key,
            self.__secondary_key: key,
        }

    def _table_name(self):
        return self._namespace

    def _table(self):
        return self._storage.Table(self._table_name())

    def _prepare_for_setter(self, data):
        return self._decode(json.dumps(data))

    def _decode(self, data):
        return json.loads(data, parse_float = decimal.Decimal)



from .adapter.abstract import NotSupported
from .adapter.inmemory import InMemory
from .mixin            import DictionaryAccess

class Storage(DictionaryAccess):
    def __init__(self, driver = None):
        self.driver = driver or InMemory()

    def get(self, key):
        return self.driver.get(key)

    def set(self, key, value):
        self.driver.set(key, value)

    def remove(self, key):
        self.driver.remove(key)

    def find(self, *args, **kwargs):
        try:
            return self.driver.find(*args, **kwargs)
        except NotSupported as e:
            return []
